from voxcell import VoxelData

gad67 = VoxelData.load_nrrd(snakemake.params.gad67)
vip = VoxelData.load_nrrd(snakemake.params.vip)
sst = VoxelData.load_nrrd(snakemake.params.sst)
pv = VoxelData.load_nrrd(snakemake.params.pv)
lamp5 = VoxelData(gad67.raw - vip.raw - sst.raw - pv.raw,
    gad67.voxel_dimensions, gad67.offset)

lamp5.save_nrrd(snakemake.output[0])



import json
from os.path import join

DATA_FOLDER = "../data/"
# Fixed 1.json (https://bbp.epfl.ch/nexus/web/neurosciencegraph/datamodels/resources/http%3A%2F%2Fbbp.epfl.ch%2Fneurosciencegraph%2Fontologies%2Fmba)
# from Dimitri https://bluebrainproject.slack.com/archives/C03G4KZJ735/p1652960553033779

name_regex = ".+, layer [1-6](\/3)?[a|b]?$"
acronym_regex = "[a-zA-Z]+[1-6](\/3)?[a|b]?$"

# key: old name, value: new name
corrections = {}
corrections["Somatomotor areas, Layer 1"] = "Somatomotor areas, layer 1"
corrections["Somatomotor areas, Layer 2/3"] = "Somatomotor areas, layer 2/3"
corrections["Somatomotor areas, Layer 5"] = "Somatomotor areas, layer 5"
corrections["Somatomotor areas, Layer 6a"] = "Somatomotor areas, layer 6a"
corrections["Somatomotor areas, Layer 6b"] = "Somatomotor areas, layer 6b"
corrections["Primary motor area, Layer 1"] = "Primary motor area, layer 1"
corrections["Primary motor area, Layer 2/3"] = "Primary motor area, layer 2/3"
corrections["Primary motor area, Layer 5"] = "Primary motor area, layer 5"
corrections["Primary motor area, Layer 6a"] = "Primary motor area, layer 6a"
corrections["Primary motor area, Layer 6b"] = "Primary motor area, layer 6b"
corrections["Anterior cingulate area, ventral part, 6a"] = "Anterior cingulate area, ventral part, layer 6a"
corrections["Anterior cingulate area, ventral part, 6b"] = "Anterior cingulate area, ventral part, layer 6b"
corrections["Ectorhinal area/Layer 1"] = "Ectorhinal area, layer 1"
corrections["Ectorhinal area/Layer 2/3"] = "Ectorhinal area, layer 2/3"
corrections["Ectorhinal area/Layer 5"] = "Ectorhinal area, layer 5"
corrections["Ectorhinal area/Layer 6a"] = "Ectorhinal area, layer 6a"
corrections["Ectorhinal area/Layer 6b"] = "Ectorhinal area, layer 6b"
corrections["Rostrolateral lateral visual area,layer 5"] = "Rostrolateral lateral visual area, layer 5" 
corrections["Laterolateral anterior visual area,layer 5"] = "Laterolateral anterior visual area, layer 5"
corrections["Mediomedial anterior visual area,layer 5"] = "Mediomedial anterior visual area, layer 5"
corrections["Mediomedial posterior visual area,layer 5"] = "Mediomedial posterior visual area, layer 5"
corrections["Medial visual area,layer 5"] = "Medial visual area, layer 5"
corrections["Rostrolateral area, layer 1"] = "Rostrolateral visual area, layer 1" 
corrections["Rostrolateral area, layer 2/3"] = "Rostrolateral visual area, layer 2/3" 
corrections["Rostrolateral area, layer 4"] = "Rostrolateral visual area, layer 4" 
corrections["Rostrolateral area, layer 5"] = "Rostrolateral visual area, layer 5" 
corrections["Rostrolateral area, layer 6a"] = "Rostrolateral visual area, layer 6a" 
corrections["Rostrolateral area, layer 6b"] = "Rostrolateral visual area, layer 6b"

jsoncontent = json.loads(open(join(DATA_FOLDER, "1.json"), "r").read())


def correct_hierarchy_names(current, corrections):
    if len(corrections) >0:
        for i, child in enumerate(current["children"]):
            current["children"][i] = correct_hierarchy_names(child, corrections)
        old_name = current["name"]
        if old_name in corrections.keys():
            current["name"] = corrections[old_name]
            del corrections[old_name]
    return current

correct_hierarchy_names(jsoncontent["msg"][0], corrections)

with open(join(DATA_FOLDER, "mba_corrected.json"), 'w') as fp:
    json.dump(jsoncontent, fp, indent=1)



import shutil
from customize_pipeline.customize_pipeline import main

# merge default output and single-region outputs into a merged file
main(snakemake.input.hierarchy, snakemake.input.annotation, snakemake.params.user_rule,
     snakemake.input.default_output_dir, snakemake.output.dir,
     snakemake.input.default_output_file, snakemake.params.metadata_path)
# replace the default output with the merged file
shutil.copytree(snakemake.output.dir, snakemake.input.default_output_dir, dirs_exist_ok=True)



from nrrdhlp.region_annotations import AnnotationWrapper

forge = None
# a "forge" argument is needed in the next line to fetch the hierarchy/annotation in case "use_{hier,anno}_file" is None
w = AnnotationWrapper(forge, method_dict={"root": "create"}, use_hier_file=snakemake.input.hierarchy, use_anno_file=snakemake.input.annotation)
w.fix(fn_hier_out=snakemake.output.hierarchy, fn_ann_out=snakemake.output.annotation, fn_log=snakemake.log[0])



import os
import json
from pathlib import Path


def create_placement_hints_metadata(ph_dir, files_ext, region_name, output_path):
    ph_files = [str(path) for path in Path(ph_dir).rglob("*" + files_ext)]
    ph_region_map = {}
    for ph_file in ph_files:
        ph_region_map[os.path.basename(ph_file)] = [region_name]

    with open(output_path, "w") as outfile:
        outfile.write(json.dumps(ph_region_map, indent=4))


shell_command = f"""{snakemake.params.app}  \
    --hierarchy-path {snakemake.input.hierarchy}  \
    --annotation-path {snakemake.input.annotation}  \
    --metadata-path {snakemake.input.region_filter}  \
    --direction-vectors-path {snakemake.input.direction_vectors}  \
    --algorithm voxel-based  \
    --output-dir {snakemake.output.dir}  \
    2>&1 | tee {snakemake.log[0]}"""
# Create Placement hints
os.system(shell_command)
create_placement_hints_metadata(snakemake.output.dir, snakemake.params.files_ext, snakemake.params.region, snakemake.output.metadata)



'''
This is in charge of pushing the OBJ region meshes
'''

import uuid
import glob
import argparse
from kgforge.core import KnowledgeGraphForge
from kgforge.core import Resource
import sys
import json
import os
import numpy as np
import nrrd
from datetime import datetime
import blue_brain_atlas_web_exporter.TreeIndexer as TreeIndexer


REGION_MESH_SCHEMAS_ID = "https://neuroshapes.org/dash/brainparcellationmesh"


def generate_payload_region_mesh(id, name, description, atlas_release_id, srs_id, region_node):
    return {
        "@id": id,
        "@type": [
            "Mesh",
            "BrainParcellationMesh"
        ],
        "atlasRelease": {
            "@id": atlas_release_id
        },
        "brainLocation": {
            "atlasSpatialReferenceSystem": {
                "@id": srs_id,
                "@type": [
                    "AtlasSpatialReferenceSystem",
                    "BrainAtlasSpatialReferenceSystem"
                ]
          },
          "brainRegion": {
              "@id": f"mba:{region_node['id']}",
              "label": region_node["name"],
          }
        },
        "description": description,
        "isRegisteredIn": {
            "@id": srs_id,
            "@type": [
                "BrainAtlasSpatialReferenceSystem",
                "AtlasSpatialReferenceSystem"
            ]
        },
        "name": name,
        "spatialUnit": "µm",
        "subject": {
            "@type": "Subject",
            "species": {
                "@id": "http://purl.obolibrary.org/obo/NCBITaxon_10090",
                "label": "Mus musculus"
            }
        }
    }




def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--forge-config",
        dest="forge_config",
        required=True,
        metavar="<FILE PATH>",
        help="The path to the Nexus Forge configuration file")

    parser.add_argument(
        "--nexus-env",
        dest="nexus_env",
        required=True,
        metavar="<URL>",
        help="URL to the Nexus environment")

    parser.add_argument(
        "--nexus-org",
        dest="nexus_org",
        required=True,
        metavar="<NAME>",
        help="Name of the Nexus organization to push data to")

    parser.add_argument(
        "--nexus-proj",
        dest="nexus_proj",
        required=True,
        metavar="<NAME>",
        help="Name of the Nexus project to push data to")

    parser.add_argument(
        "--access-token",
        dest="access_token",
        required=True,
        metavar="<NAME>",
        help="Access token (JWT) to push data to Nexus")
        
    parser.add_argument(
        "--nexus-id-aibs-ccf-srs",
        dest="nexus_id_aibs_ccf_srs",
        required=True,
        metavar="<NEXUS ID>",
        help="Nexus ID for the Spatial Reference System")

    parser.add_argument(
        "--hierarchy",
        dest="hierarchy",
        required=True,
        metavar="<FILE PATH>",
        help="The hierarchy JSON file, sometimes called 1.json")

    parser.add_argument(
        "--region-mesh-dir",
        dest="region_mesh_dir",
        required=True,
        metavar="<DIR PATH>",
        help="The directory containing the NRRD region masks")

    parser.add_argument(
        "--atlasrelease-id",
        dest="atlasrelease_id",
        required=False,
        metavar="<NEXUS ID>",
        help="Nexus ID of the AtlasRelease Nexus Resource the volumes are linked to")

    return parser.parse_args()



def main():
    args = parse_args()

    forge = KnowledgeGraphForge(
        args.forge_config,
        endpoint=args.nexus_env,
        bucket=f"{args.nexus_org}/{args.nexus_proj}",
        token=args.access_token,
        debug=True,
    )

    forge._debug=True

    srs_id = args.nexus_id_aibs_ccf_srs
    atlas_release_id = args.atlasrelease_id

    # Reading the brain region ontology, to later being able to generate
    # better names that include brain region labels
    brain_onto_json = json.loads(open(args.hierarchy, "r").read())
    # sometimes, the 1.json has its content in a "msg" sub prop (the original version has).
    # and some other versions don't. Here we deal with both
    if "msg" in brain_onto_json:
        onto_flat_tree = TreeIndexer.flattenTree(brain_onto_json['msg'][0])
    else:
        onto_flat_tree = TreeIndexer.flattenTree(brain_onto_json)

    # finding all the masks
    mesh_paths = glob.glob(f"{args.region_mesh_dir}/*.obj")

    print("Pushing to Nexus: region meshes...")
    for i in range(0, len(mesh_paths)):
        mesh_filepath = mesh_paths[i]
        region_id = int(os.path.basename(mesh_filepath).split(".")[0])
        region_node = onto_flat_tree[region_id]
        region_mesh_nexus_id = forge.format("identifier", str(uuid.uuid4()))
        print(f"[{i+1}/{len(mesh_paths)}] ", region_node["name"], " --> ", region_mesh_nexus_id)

        region_mesh_payload = generate_payload_region_mesh(region_mesh_nexus_id,
            f"Mesh of {region_node['name']}",
            f"Blue Brain Atlas mesh of the region {region_node['name']}",
            atlas_release_id,
            srs_id,
            region_node
        )

        region_mask_volume_resource = Resource.from_json(region_mesh_payload)
        region_mask_volume_resource.distribution = forge.attach(mesh_filepath, content_type="application/obj")
        forge.register(region_mask_volume_resource, schema_id=REGION_MESH_SCHEMAS_ID)

if __name__ == "__main__":
    main()



import os
import yaml

import gitlab
from kgforge.core import KnowledgeGraphForge

nexus_token = os.environ["NEXUS_TOKEN"]
resource_tag = os.environ["RESOURCE_TAG"]
bmo_repo_token = os.environ["BMO_ACCESS_TOKEN"]
is_prod_env = os.environ["IS_PROD_ENV"]
if is_prod_env:
    endpoint = "https://bbp.epfl.ch/nexus/v1"
    resource_id = os.environ["ATLAS_PARCELLATION_ONTOLOGY_ID_PROD"]
    target_branch = "main"
    ci_var_name = 'ATLAS_PARCELLATION_ONTOLOGY_VERSION_PROD'
    new_branch = "update_atlas_parcellation_ontology_prod"
else:
    endpoint = "https://staging.nise.bbp.epfl.ch/nexus/v1"
    resource_id = os.environ["ATLAS_PARCELLATION_ONTOLOGY_ID_STAGING"]
    target_branch = "develop"
    ci_var_name = 'ATLAS_PARCELLATION_ONTOLOGY_VERSION_STAGING'
    new_branch = "update_atlas_parcellation_ontology_staging"

forge = KnowledgeGraphForge("forge-config.yml", bucket="bbp/atlas",
    endpoint=endpoint, token=nexus_token)
resource = forge.retrieve(id=resource_id, version=resource_tag)
if not resource:
    raise Exception(f"Tag '{resource_tag}' does not exist for Resource Id {resource_id}"
                    f" in bucket 'bbp/atlas' (endpoint '{endpoint}')")

resource_rev = resource._store_metadata._rev
gl = gitlab.Gitlab('https://bbpgitlab.epfl.ch', private_token=bmo_repo_token)
project = gl.projects.get("dke/apps/brain-modeling-ontology")
gl_ci = project.files.get(file_path='.gitlab-ci.yml', ref=target_branch)
gl_ci_dict = yaml.safe_load(gl_ci.decode())
ci_var_value = gl_ci_dict['variables'][ci_var_name]
if ci_var_value == resource_rev:
    raise Exception(f"The current value of {ci_var_name} in branch '{target_branch}' is"
                    f" already {resource_rev}, nothing to update")

print(f"Updating {ci_var_name} from {ci_var_value} to {resource_rev}")
if new_branch not in project.branches.list():
    project.branches.create({'branch': new_branch, 'ref': target_branch})
raw_content = gl_ci.decode()
gl_ci.content = raw_content.decode().replace(f"{ci_var_name}: {ci_var_value}", f"{ci_var_name}: {resource_rev}")
gl_ci.save(branch=new_branch, commit_message=f'Update {ci_var_name} to {resource_rev}')

project.mergerequests.create({
    'source_branch': new_branch,
    'target_branch': target_branch,
    'title': f'Update {ci_var_name}'
})



'''
This is in charge of pushing the NRRD region masks

'''

import uuid
import glob
import argparse
from kgforge.core import KnowledgeGraphForge
from kgforge.core import Resource
import sys
import json
import os
import numpy as np
import nrrd
from datetime import datetime
import blue_brain_atlas_web_exporter.TreeIndexer as TreeIndexer



NRRD_TYPES_TO_NUMPY = {
    "signed char": "int8",
    "int8": "int8",
    "int8_t": "int8",
    "uchar": "uint8",
    "unsigned char": "uint8",
    "uint8": "uint8",
    "uint8_t": "uint8",
    "short": "int16",
    "short int": "int16",
    "signed short": "int16",
    "signed short int": "int16",
    "int16": "int16",
    "int16_t": "int16",
    "ushort": "int16",
    "unsigned short": "uint16",
    "unsigned short int": "uint16",
    "uint16": "uint16",
    "uint16_t": "uint16",
    "int": "int32",
    "signed int": "int32",
    "int32": "int32",
    "int32_t": "int32",
    "uint": "uint32",
    "unsigned int": "uint32",
    "uint32": "uint32",
    "uint32_t": "uint32",
    "longlong": "int64",
    "long long": "int64",
    "long long int": "int64",
    "signed long long": "int64",
    "signed long long int": "int64",
    "int64": "int64",
    "int64_t": "int64",
    "ulonglong": "uint64",
    "unsigned long long": "uint64",
    "unsigned long long int": "uint64",
    "uint64": "uint64",
    "uint64_t": "uint64",
    "float": "float32",
    "double": "float64"
}


dataSampleModalities = {
  "PARCELLATION_ID": "parcellationId",
  "CELL_TYPE_ID": "cellTypeId",
  "COLOR_RGB": "colorRGB",
  "COLOR_RGBA": "colorRGBA",
  "EULER_ANGLE": "eulerAngle",
  "GRADIENT_2D": "gradient2D",
  "GRADIENT_3D": "gradient3D",
  "LUMINANCE": "luminance",
  "MARKER_INTENSITY": "markerIntensity",
  "MASK": "mask",
  "POSITION_2D": "position2D",
  "POSITION_3D": "position3D",
  "QUANTITY": "quantity",
  "QUATERNION": "quaternion",
}

dataSampleModalitiesNumberOfComponents = {
  "PARCELLATION_ID": 1,
  "CELL_TYPE_ID": 1,
  "COLOR_RGB": 3,
  "COLOR_RGBA": 4,
  "EULER_ANGLE": 3,
  "GRADIENT_2D": 2,
  "GRADIENT_3D": 3,
  "LUMINANCE": 1,
  "MARKER_INTENSITY": 1,
  "MASK": 1,
  "POSITION_2D": 2,
  "POSITION_3D": 3,
  "QUANTITY": 1,
  "QUATERNION": 4,
}

VOLUMETRICDATALAYER_SCHEMAS_ID = "https://neuroshapes.org/dash/volumetricdatalayer"

def generate_payload_volumetric_data_layer(filepath, id, name, description, type, atlas_release_id, srs_id, region_node, sample_modality, resolution):
    
    nrrd_header = nrrd.read_header(filepath)
    return {
        "@id": id,
        "@type": [
            "VolumetricDataLayer",
            type,
            "Dataset"
        ],
        "atlasRelease": {
            "@id": atlas_release_id
        },
        "brainLocation": {
            "atlasSpatialReferenceSystem": {
            "@id": srs_id,
            "@type": [
                "AtlasSpatialReferenceSystem",
                "BrainAtlasSpatialReferenceSystem"
            ]
            },
            "brainRegion": {
            "@id": f"mba:{region_node['id']}",
            "label": region_node["name"],
            }
        },
        
        "bufferEncoding": nrrd_header["encoding"],

        "componentEncoding": NRRD_TYPES_TO_NUMPY[nrrd_header["type"]],

        "contribution": {
            "@type": "Contribution",
            "agent": {
            "@id": "https://ror.org/02s376052",
            "@type": [
                "Organization",
                "Agent"
            ]
            }
        },

        "dataSampleModality": dataSampleModalities[sample_modality],

        "description": description,
        "dimension": [
            {
            "@type": "ComponentDimension",
            "name": dataSampleModalities[sample_modality],
            "size": dataSampleModalitiesNumberOfComponents[sample_modality],
            },
            {
            "@type": "SpaceDimension",
            "size": 528,
            "unitCode": "voxel"
            },
            {
            "@type": "SpaceDimension",
            "size": 320,
            "unitCode": "voxel"
            },
            {
            "@type": "SpaceDimension",
            "size": 456,
            "unitCode": "voxel"
            }
        ],
   
        "endianness": nrrd_header.get("endian", "little"),

        "fileExtension": "nrrd",
        "isRegisteredIn": {
            "@id": srs_id,
            "@type": [
            "AtlasSpatialReferenceSystem",
            "BrainAtlasSpatialReferenceSystem"
            ]
        },
        "name": name,
        "resolution": {
            "unitCode": "µm",
            "value": resolution
        },
        "sampleType": "label",
        "subject": {
            "@type": "Subject",
            "species": {
            "@id": "http://purl.obolibrary.org/obo/NCBITaxon_10090",
            "label": "Mus musculus"
            }
        },
        "worldMatrix": [
            resolution,
            0,
            0,
            0,
            0,
            resolution,
            0,
            0,
            0,
            0,
            resolution,
            0,
            0,
            0,
            0,
            1
        ]
        }


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--forge-config",
        dest="forge_config",
        required=True,
        metavar="<FILE PATH>",
        help="The path to the Nexus Forge configuration file")

    parser.add_argument(
        "--nexus-env",
        dest="nexus_env",
        required=True,
        metavar="<URL>",
        help="URL to the Nexus environment")

    parser.add_argument(
        "--nexus-org",
        dest="nexus_org",
        required=True,
        metavar="<NAME>",
        help="Name of the Nexus organization to push data to")

    parser.add_argument(
        "--nexus-proj",
        dest="nexus_proj",
        required=True,
        metavar="<NAME>",
        help="Name of the Nexus project to push data to")

    parser.add_argument(
        "--access-token",
        dest="access_token",
        required=True,
        metavar="<NAME>",
        help="Access token (JWT) to push data to Nexus")
        
    parser.add_argument(
        "--nexus-id-aibs-ccf-srs",
        dest="nexus_id_aibs_ccf_srs",
        required=True,
        metavar="<NEXUS ID>",
        help="Nexus ID for the Spatial Reference System")

    parser.add_argument(
        "--hierarchy",
        dest="hierarchy",
        required=True,
        metavar="<FILE PATH>",
        help="The hierarchy JSON file, sometimes called 1.json")

    parser.add_argument(
        "--region-mask-volume-dir",
        dest="region_mask_volume_dir",
        required=True,
        metavar="<DIR PATH>",
        help="The directory containing the NRRD region masks")

    parser.add_argument(
        "--atlasrelease-id",
        dest="atlasrelease_id",
        required=False,
        metavar="<NEXUS ID>",
        help="Nexus ID of the AtlasRelease Nexus Resource the volumes are linked to")

    return parser.parse_args()



def main():
    args = parse_args()

    forge = KnowledgeGraphForge(
        args.forge_config,
        endpoint=args.nexus_env,
        bucket=f"{args.nexus_org}/{args.nexus_proj}",
        token=args.access_token,
        debug=True,
    )

    forge._debug=True

    srs_id = args.nexus_id_aibs_ccf_srs
    atlas_release_id = args.atlasrelease_id

    # Reading the brain region ontology, to later being able to generate
    # better names that include brain region labels
    brain_onto_json = json.loads(open(args.hierarchy, "r").read())
    # sometimes, the 1.json has its content in a "msg" sub prop (the original version has).
    # and some other versions don't. Here we deal with both
    if "msg" in brain_onto_json:
        onto_flat_tree = TreeIndexer.flattenTree(brain_onto_json['msg'][0])
    else:
        onto_flat_tree = TreeIndexer.flattenTree(brain_onto_json)

    # finding all the masks
    masks_nrrd_paths = glob.glob(f"{args.region_mask_volume_dir}/*.nrrd")

    print("Pushing to Nexus: region masks volume...")
    for i in range(0, len(masks_nrrd_paths)):
        mask_filepath = masks_nrrd_paths[i]
        region_id = int(os.path.basename(mask_filepath).split(".")[0])
        region_node = onto_flat_tree[region_id]
        region_mask_nexus_id = forge.format("identifier", str(uuid.uuid4()))
        print(f"[{i+1}/{len(masks_nrrd_paths)}] ", region_node["name"], " --> ", region_mask_nexus_id)

        region_mask_volume_payload = generate_payload_volumetric_data_layer(mask_filepath, 
            region_mask_nexus_id,
            f"Mask Volume of {region_node['name']}, 25µm",
            f"Blue Brain Atlas mask raster volume of the region {region_node['name']}, at resolution 25µm",
            "BrainParcellationMask",
            atlas_release_id,
            srs_id,
            region_node,
            "MASK",
            25,
        )
        region_mask_volume_resource = Resource.from_json(region_mask_volume_payload)
        region_mask_volume_resource.distribution = forge.attach(mask_filepath, content_type="application/nrrd")
        forge.register(region_mask_volume_resource, schema_id=VOLUMETRICDATALAYER_SCHEMAS_ID)

if __name__ == "__main__":
    main()



import os
import json
from pathlib import Path
from copy import deepcopy
import numpy as np

from voxcell import RegionMap, VoxelData


def main(hierarchy, annotation_volume, user_rule, default_rule_output,
         merged_output_dir, default_rule_file=None, metadata_path=None):
    region_map = RegionMap.load_json(hierarchy)
    annotation = VoxelData.load_nrrd(annotation_volume)

    rule_name = user_rule["rule"]

    region_volume_map = {}
    customized_regions = user_rule["execute"]
    for custom_region in customized_regions:
        region_id = get_region_id(custom_region["brainRegion"])
        region_volume_map[region_id] = custom_region["output_dir"]

    print(f"Merging outputs of rule {rule_name} from {len(customized_regions)} regions")
    merged_volumes = merge_nrrd_files(region_map, annotation.raw, region_volume_map,
        default_rule_output, merged_output_dir, default_rule_file, metadata_path)
    print(f"{len(merged_volumes)} files have been merged in {merged_output_dir}: {merged_volumes}")


def merge_nrrd_files(region_map: RegionMap, annotation: np.ndarray,
    region_volume_map: dict, default_rule_output: str, merged_output_dir: str,
    default_rule_file=None, metadata_path=None) -> list:
    """
    Merge nrrd volumes for various brain regions.

    Args:
        region_map: voxcell.RegionMap of the regions hierarchy
        annotation: annotation volume, where each voxel contains a region id
        region_volume_map: mapping between brain region and its corresponding nrrd file to merge
        default_rule_output: output path of the nrrd file of original default rule.
            The areas of the brain regions in region_volume_map will be superseded.
        merged_output_dir: directory where to save merged volumes
        metadata_path: optional path to the metadata file

    Returns:
        The list of volume files with updated values from the volumes in region_volume_map.
    """

    extension = ".nrrd"
    default_output_files = []
    if not os.path.exists(default_rule_output):
        raise Exception("The output of the default rule does not exist at", default_rule_output)
    if os.path.isdir(default_rule_output):
        if default_rule_file:  # TODO account for list of default_rule_files
            print(f"A default output filepath is provided: {default_rule_file}, "
                  "only such a file will be merged from each region-specific output dir.")
            glob_str = os.path.basename(default_rule_file)
        else:
            glob_str = "*" + extension
        default_output_files.extend([str(path) for path in Path(default_rule_output).glob(glob_str)])
    elif default_rule_output.endswith(extension):
        if os.path.isfile(default_rule_output):
            default_output_files = [default_rule_output]

    if metadata_path:
        metadata_file = open(metadata_path, "r+")
        metadata_json = json.load(metadata_file)

    result = []
    os.makedirs(merged_output_dir, exist_ok=True)
    for default_output in default_output_files:
        filename = os.path.basename(default_output)
        # Get the default volume
        default_volume = VoxelData.load_nrrd(default_output)
        result_volume = np.copy(default_volume.raw)
        # Update the result regions with values from the input map
        for (region_id, volume_path) in region_volume_map.items():
            ids_reg = region_map.find(region_id, "id", with_descendants=True)
            if not ids_reg:
                print(f"Warning: region {region_id} is not found in the hierarchy provided")
                continue

            volume_file = os.path.join(volume_path, filename)
            if not os.path.isfile(volume_file):
                print(f"Warning: no file {filename} found in {volume_path}, skipping it.")
                continue
            volume = VoxelData.load_nrrd(volume_file).raw

            # Get region mask
            region_mask = np.isin(annotation, list(ids_reg))
            # Supersede region {region_id} in result with values from volume
            if result_volume.shape[0:len(region_mask.shape)] == region_mask.shape:
                result_volume[region_mask] = volume[region_mask]
            else:
                result_volume[:region_mask] = volume[:region_mask]

            if metadata_path:
                metadata_json[filename].append(region_map.get(region_id, "name"))

        merged_file = os.path.join(merged_output_dir, filename)
        default_volume.with_data(result_volume).save_nrrd(merged_file)
        result.append(merged_file)

    if metadata_path:
        metadata_file.seek(0)
        json.dump(metadata_json, metadata_file)
        metadata_file.truncate()
        metadata_file.close()

    return result


def get_region_id(full_id):
    parts = full_id.split("/")
    return int(parts[-1])


def check_rule_existence(rule, rule_name):
    if not rule:
        raise Exception(f"Rule '{rule_name}' does not exist in the default pipeline. \
            The existing rules can be listed with the command:\nsnakemake --list")


def get_merge_rule_name(rule_name):
    return "merge_" + rule_name


def get_var_path_map(available_vars, pipeline_dataset_config):
    var_path_map = {}
    for (var, value) in available_vars.items():
        steps = value.split(":")
        path = deepcopy(pipeline_dataset_config)
        for step in steps:
            path = path[step]
        var_path_map[var] = path

    return var_path_map



'''
This is in charge of pushing the OBJ region meshes
'''

import uuid
import glob
import argparse
from kgforge.core import KnowledgeGraphForge
from kgforge.core import Resource
import sys
import json
import os
import numpy as np
import nrrd
from datetime import datetime
import blue_brain_atlas_web_exporter.TreeIndexer as TreeIndexer


def generate_payload_region_summary(id, name, description, atlas_release_id, srs_id, region_node, metadata):
    return {
        "@id": id,

        "@type": [
            "RegionSummary",
            "Entity"
        ],

        "name": name,

        "description": description,

        "brainLocation": {
            "atlasSpatialReferenceSystem": {
            "@id":srs_id,
            "@type": [
                "BrainAtlasSpatialReferenceSystem",
                "AtlasSpatialReferenceSystem"
            ]
            },
            "brainRegion": {
                "@id": f"mba:{region_node['id']}",
                "label": region_node["name"],
            }
        },

        "volume": {
            "ratio": metadata["regionVolumeRatioToWholeBrain"],
            "total": {
                "size": metadata["regionVolume"],
                "unitCode": "µm³"
            }
        },

        "adjacentTo": list(map(lambda region_id: { "@id": f"mba:{region_id}", "ratio": metadata["adjacentTo"][region_id]}, list(metadata["adjacentTo"].keys()))),

        "continuousWith": list(map(lambda region_id: { "@id": f"mba:{region_id}" }, metadata["continuousWith"])),

        "layers": metadata["layers"],



        "atlasRelease": {
            "@id": atlas_release_id,
            "@type": [
            "AtlasRelease",
            "BrainAtlasRelease",
            "Entity"
            ]
        }
    }


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--forge-config",
        dest="forge_config",
        required=True,
        metavar="<FILE PATH>",
        help="The path to the Nexus Forge configuration file")

    parser.add_argument(
        "--nexus-env",
        dest="nexus_env",
        required=True,
        metavar="<URL>",
        help="URL to the Nexus environment")

    parser.add_argument(
        "--nexus-org",
        dest="nexus_org",
        required=True,
        metavar="<NAME>",
        help="Name of the Nexus organization to push data to")

    parser.add_argument(
        "--nexus-proj",
        dest="nexus_proj",
        required=True,
        metavar="<NAME>",
        help="Name of the Nexus project to push data to")

    parser.add_argument(
        "--access-token",
        dest="access_token",
        required=True,
        metavar="<NAME>",
        help="Access token (JWT) to push data to Nexus")
        
    parser.add_argument(
        "--nexus-id-aibs-ccf-srs",
        dest="nexus_id_aibs_ccf_srs",
        required=True,
        metavar="<NEXUS ID>",
        help="Nexus ID for the Spatial Reference System")

    parser.add_argument(
        "--hierarchy",
        dest="hierarchy",
        required=True,
        metavar="<FILE PATH>",
        help="The hierarchy JSON file, sometimes called 1.json")

    parser.add_argument(
        "--region-metadata",
        dest="region_metadata",
        required=True,
        metavar="<FILE PATH>",
        help="The JSON file with brain region metadata")

    parser.add_argument(
        "--atlasrelease-id",
        dest="atlasrelease_id",
        required=False,
        metavar="<NEXUS ID>",
        help="Nexus ID of the AtlasRelease Nexus Resource the volumes are linked to")

    return parser.parse_args()



def main():
    args = parse_args()

    forge = KnowledgeGraphForge(
        args.forge_config,
        endpoint=args.nexus_env,
        bucket=f"{args.nexus_org}/{args.nexus_proj}",
        token=args.access_token,
        debug=True,
    )

    forge._debug=True

    srs_id = args.nexus_id_aibs_ccf_srs
    atlas_release_id = args.atlasrelease_id

    # Reading the brain region ontology, to later being able to generate
    # better names that include brain region labels
    brain_onto_json = json.loads(open(args.hierarchy, "r").read())
    # sometimes, the 1.json has its content in a "msg" sub prop (the original version has).
    # and some other versions don't. Here we deal with both
    if "msg" in brain_onto_json:
        onto_flat_tree = TreeIndexer.flattenTree(brain_onto_json['msg'][0])
    else:
        onto_flat_tree = TreeIndexer.flattenTree(brain_onto_json)

    all_regions_meta = json.loads(open(args.region_metadata).read())
    region_keys = list(all_regions_meta.keys())

    for i in range(0, len(region_keys)):
        region_id = region_keys[i]
        region_node = onto_flat_tree[int(region_id)]
        summary_metadata = all_regions_meta[region_id]
        region_summary_nexus_id = forge.format("identifier", str(uuid.uuid4()))

        print(f"[{i+1}/{len(region_keys)}] ", region_node["name"], " --> ", region_summary_nexus_id)

        region_summary_payload = generate_payload_region_summary(region_summary_nexus_id,
            f"Region Summary of {region_node['name']}",
            f"Some details about the region {region_node['name']} as in the Blue Brain Atlas, regarding volume and neighbouring",
            atlas_release_id,
            srs_id,
            region_node,
            summary_metadata
        )

        region_summary_resource = Resource.from_json(region_summary_payload)
        forge.register(region_summary_resource)


if __name__ == "__main__":
    main()



import os
import click
import json
import yaml
import re


@click.command()
@click.option("--target-rule",
              type=click.STRING, required=True,
              help="The target rule of the pipeline to execute")
@click.option("--user-config-path",
              type=click.Path(exists=True), required=False,
              help="The user configuration to customize the pipeline")
@click.option("--repo-path",
              type=click.Path(exists=True), required=False, default=".",
              help="The path of the pipeline repository")
@click.option("--snakemake-options",
              type=click.STRING, required=False, default="",
              help="String of options to pass to the snakemake command")
@click.option("--service-token",
              is_flag=True, default=False,
              help="Flag to use a service token instead of a user token",
              show_default=True)
@click.option("--token-username",
              type=click.STRING, required=False,
              help="Username for token fetcher")
@click.option("--token-password",
              type=click.STRING, required=False,
              help="Password for token fetcher")
def execute_pipeline(target_rule, user_config_path, repo_path, snakemake_options,
                     service_token, token_username, token_password):
    options_separator = "  "
    if user_config_path:
        if "--snakefile " in snakemake_options:
            raise Exception("The '--snakefile' option can not be used together with --user-config-path")

        from blue_brain_token_fetch.token_fetcher_user import TokenFetcherUser
        from blue_brain_token_fetch.token_fetcher_service import TokenFetcherService
        from pipeline_validator.pipeline_validator import pipeline_validator
        from customize_pipeline.customize_pipeline import get_merge_rule_name, get_var_path_map

        with open(os.path.join(repo_path, "config.yaml")) as pipeline_config_file:
            pipeline_config = yaml.safe_load(pipeline_config_file.read().strip())
        keycloak_config = os.path.join(repo_path, pipeline_config["KEYCLOAK_CONFIG"])
        working_dir = pipeline_config["WORKING_DIR"]
        if not service_token:
            token_fetcher = TokenFetcherUser(token_username, token_password, keycloak_config_file=keycloak_config)
        else:
            token_fetcher = TokenFetcherService(token_username, token_password, keycloak_config_file=keycloak_config)

        with open(os.path.join(repo_path, "rules_config_dir_templates/push_dataset_config_template.yaml"), "r") as push_dataset_config_template:
            push_dataset_config = re.sub("{WORKING_DIR}", working_dir, push_dataset_config_template.read())
            push_dataset_config_dict = yaml.safe_load(push_dataset_config.strip())
        with open(os.path.join(repo_path, "customize_pipeline/available_vars.yaml"), "r") as vars_file:
            available_vars = yaml.safe_load(vars_file.read().strip())
        input_group = "input"
        var_path_map = get_var_path_map(available_vars[input_group], push_dataset_config_dict)
        whitelisted_vars = [f"{input_group}.{var}" for var in var_path_map.keys()]

        pipeline_validator(user_config_path, token_fetcher.get_access_token(), whitelisted_vars)

        user_config_json = json.load(open(user_config_path))
        priority_rules = []
        for user_rule_name in [user_rule["rule"] for user_rule in user_config_json["rules"]]:
            merge_rule_name = get_merge_rule_name(user_rule_name)
            if target_rule == user_rule_name:
                target_rule = merge_rule_name
            else:
                priority_rules.append(merge_rule_name)

        snakemake_options += f" --snakefile {repo_path}/customize_pipeline/custom_snakefile"
        if priority_rules:
            # this should not be the last option otherwise target_rule gets prioritized too
            snakemake_options += f" --prioritize {' '.join(priority_rules)}"

        snakemake_config_flag = "--config "
        user_config_option = f"USER_CONFIG={user_config_path}"
        target_config_option = f"TARGET_RULE={target_rule}"
        repo_path_option = f"REPO_PATH={repo_path}"
        full_config_option = f"{snakemake_config_flag} {user_config_option} {repo_path_option} {target_config_option}"
        if snakemake_config_flag in snakemake_options:
            snakemake_options = snakemake_options.replace(snakemake_config_flag,
                full_config_option + options_separator)
        else:
            snakemake_options = options_separator.join([full_config_option, snakemake_options])

    full_snakemake_options = "--printshellcmds"
    if snakemake_options:
        full_snakemake_options = options_separator.join([snakemake_options, full_snakemake_options])

    pipeline_command = options_separator.join(["snakemake", full_snakemake_options, target_rule])
    print("\nExecuting command:\n", pipeline_command)
    os.system(pipeline_command)



import os

mesh_dir_option = ""
if snakemake.params.export_meshes:
    mesh_dir_option = f" --out-mesh-dir {snakemake.output.mesh_dir}"
else:
    os.makedirs(snakemake.output.mesh_dir, exist_ok=True)

shell_command = f"""{snakemake.params.app}  \
    --hierarchy {snakemake.input.hierarchy}  \
    --parcellation-volume {snakemake.input.annotation}  \
    --region-layer-map {snakemake.input.region_layer_map}  \
    {mesh_dir_option}  \
    --out-mask-dir {snakemake.output.mask_dir}  \
    --out-metadata {snakemake.output.json_metadata_parcellations}  \
    --out-hierarchy-volume {snakemake.output.hierarchy_volume}  \
    --out-hierarchy-jsonld {snakemake.output.hierarchy_jsonld}  \
    2>&1 | tee {snakemake.log[0]}"""

os.system(shell_command)



import os
import logging

from kgforge.core import KnowledgeGraphForge
from cellCompVolume_payload import create_payload

from blue_cwl.validation import validate_schema

logging.basicConfig(level=logging.INFO)
L = logging.getLogger(__name__)

forge_config = "forge-config.yml" 
nexus_env = "https://staging.nise.bbp.epfl.ch/nexus/v1"
nexus_org = "bbp"
nexus_proj = "atlas"
nexus_token = os.environ["NEXUS_STAGING_TOKEN"]
test_folder = os.environ["TEST_FOLDER"]
tag = "v0.5.1-rc1"
expected_densities = 245
atlasrelease_id = "https://bbp.epfl.ch/neurosciencegraph/data/brainatlasrelease/c96c71a8-4c0d-4bc1-8a1a-141d9ed6693d"


def test_cellCompVolume_payload():
    bucket = "/".join([nexus_org, nexus_proj])
    forge = KnowledgeGraphForge(forge_config, bucket=bucket, endpoint=nexus_env, token=nexus_token)
                       
    output_file = os.path.join(test_folder, "cellCompositionVolume_payload.json")
    payload = create_payload(forge, atlasrelease_id, output_file, expected_densities,
        endpoint=nexus_env, bucket=bucket, tag=tag)
    L.info("Test output: %s" % output_file)

    validate_schema(data=payload, schema_name="cell_composition_volume_distribution.yml")



import os
import logging
import json
from pathlib import Path

from voxcell import RegionMap
from kgforge.core import KnowledgeGraphForge

logging.basicConfig(level=logging.INFO)
L = logging.getLogger(__name__)

forge_config = "forge-config.yml" 
nexus_env = "https://staging.nise.bbp.epfl.ch/nexus/v1"
nexus_org = "bbp"
nexus_proj = "atlas"
nexus_token = os.environ["NEXUS_STAGING_TOKEN"]
test_folder = os.environ["TEST_FOLDER"]

metadata_dir = "metadata"


def test_metadata():
    prob_maps = [str(path) for path in Path(metadata_dir).rglob("probability_map_*.csv")]
    prob_maps_n = len(prob_maps)
    if prob_maps_n < 1:
        return

    L.info(f"Testing labels from {prob_maps_n} probability maps in {metadata_dir}/: {', '.join(prob_maps)}")
    forge = KnowledgeGraphForge(forge_config, bucket="/".join([nexus_org, nexus_proj]),
                                endpoint=nexus_env, token=nexus_token)
    nexus_endpoint = f"Nexus endpoint: '{nexus_env}"

    me_separator = "|"
    types_to_resolve = set()
    for prob_map_file in prob_maps:
        with open(prob_map_file) as prob_map:
            header = prob_map.readline().strip('\n')
            columns = header.split(",")
            me_types = [column for column in columns if me_separator in column]
            for me_type in me_types:
                types_to_resolve.update(me_type.split(me_separator))

    for type_to_resolve in types_to_resolve:
        res = forge.resolve(type_to_resolve, scope="ontology", target="CellType", strategy="EXACT_MATCH")
        assert res, f"Label '{type_to_resolve}' is not resolved from {nexus_endpoint}"

    L.info(f"All M and E type labels in {metadata_dir} can be resolved from {nexus_endpoint}")


def test_regions_to_layers_map_and_PH_map():
    forge = KnowledgeGraphForge(forge_config, bucket="/".join(["neurosciencegraph", "datamodels"]),
                                endpoint=nexus_env, token=nexus_token)

    layers_regions_map_path = os.path.join(metadata_dir, "PH_layers_regions_map.json")
    L.info(f"Testing the layer labels in the layers to regions map from {layers_regions_map_path}")
    with open(layers_regions_map_path) as layers_regions_map_file:
        layers_regions_map = json.load(layers_regions_map_file)
    for filename in layers_regions_map:
        for region, layer in layers_regions_map[filename].items():
            check_region(region)
            check_label(layer, forge, region, layers_regions_map_path)

    regions_layers_map_path = os.path.join(metadata_dir, "regions_layers_map.json")
    L.info(f"Testing the layer labels in the layers to regions map from {regions_layers_map_path}")
    with open(regions_layers_map_path) as regions_layers_map_file:
        regions_layers_map = json.load(regions_layers_map_file)
    for region in regions_layers_map:
        for layer in regions_layers_map[region]["layers"]:
            check_label(layer, forge, region, regions_layers_map_path)


def check_region(region_acronym):
    test_data = os.path.join(test_folder, "data")
    hierarchy_file = os.path.join(test_data, "hierarchy_leaves_only.json")
    region_map = RegionMap.load_json(hierarchy_file)

    id_reg = region_map.find(region_acronym, "acronym")
    if not id_reg:
        raise Exception(f"Region acronym '{region_acronym}' is not found in the hierarchy.")


forge_retrieve_cache = {}

def check_label(layer, forge, region, file_path):
    layer_id = layer["layer_ID"]
    layer_label = layer["layer_label"]
    if layer_label not in forge_retrieve_cache:
        layer_res = forge.retrieve(layer_id)
        if not layer_res:
            raise Exception(f"Resource ID {layer_id} can not be retrieved from bucket '{forge._store.bucket}'")
        forge_retrieve_cache[layer_label] = layer_res
    else:
        layer_res = forge_retrieve_cache[layer_label]

    layer_res_label = layer_res.label
    assert layer_res_label == layer_label, (f"The label '{layer_label}' of layer_ID '{layer_id}'"
        f" for region {region} in {file_path} does not correspond to the resolved label '{layer_res_label}'")



import os
import json
import yaml
from pathlib import Path
import hashlib

from kgforge.core import KnowledgeGraphForge
from kgforge.core import Resource

nexus_token = os.environ["NEXUS_TOKEN"]
nexus_token_staging = os.environ["NEXUS_STAGING_TOKEN"]
nexus_ids_path = os.environ["NEXUS_IDS_PATH"]
res_tag_sep = "?tag="
metadata_dir = os.environ["METADATA_DIR"]
file_nexus_id_map = os.environ["FILE_NEXUS_ID_MAP"]
commit_sha = os.environ["COMMIT_SHA"]
is_prod_env = os.environ["IS_PROD_ENV"]


def create_prob_map_resource(name, description):
    res = Resource(
        name=name,
        description=description,
        type=["Dataset", "PipelineConfigResource"],
        generation=Resource(
            type="Generation",
            activity=Resource(
                id=commit_sha,
                type="Activity")
        )
    )
    return res


def increment_minor(res_tag):
    # tag version: "vM.m.p" (vMajor.minor.patch)
    vMajor, minor, patch = res_tag.split(".")
    new_tag = res_tag.replace(f"{vMajor}.{minor}.{patch}", f"{vMajor}.{int(minor) +1}.0")
    return new_tag


def synch_nexus():
    with open(nexus_ids_path, 'r') as nexus_ids_file:
        nexus_ids = json.loads(nexus_ids_file.read().strip())

    pipeline_config = yaml.safe_load(open("config.yaml").read().strip())

    if is_prod_env:
        forge = KnowledgeGraphForge("forge-config.yml", bucket="bbp/atlas",
                                    endpoint=pipeline_config["NEXUS_PROD_ENV"],
                                    token=nexus_token)
        synch_nexus_prod(nexus_ids, forge, pipeline_config["NEXUS_STAGING_ENV"])
    else:
        forge = KnowledgeGraphForge("forge-config.yml", bucket="bbp/atlas",
                                    endpoint=pipeline_config["NEXUS_STAGING_ENV"],
                                    token=nexus_token)
        update_nexus_staging(nexus_ids, forge)


def synch_nexus_prod(nexus_ids, forge_prod, staging_env):
    skip_synch = ["AtlasRelease", "CellComposition", "species", "ParcellationOntology", "10"]
    for res_type, res_names in nexus_ids.items():
        if res_type in skip_synch:
            continue

        if isinstance(res_names, str):
            res_id = res_names
            synch_resource(res_id, forge_prod, res_names, staging_env)
            continue

        for res_name in res_names:
            if isinstance(res_names[res_name], str):
                res_id = res_names[res_name]
                synch_resource(res_id, forge_prod, res_name, staging_env)
            else:
                resolution = res_name
                if resolution in skip_synch:
                    continue
                for res_sub_type, res_sub_names in res_names[resolution].items():
                    for res_sub_name, res_id in res_sub_names.items():
                        synch_resource(res_id, forge_prod, res_sub_name, staging_env)


def synch_resource(res_id_tag, forge_prod, res_name, staging_env):
    res_tag = None
    if res_tag_sep in res_id_tag:
        res_id, res_tag = res_id_tag.split(res_tag_sep)
    else:
        res_id = res_id_tag

    if not res_tag:
        print(f"\nResource {res_name} (Nexus id: '{res_id}') has no tag, skipping it.")
        return

    nexus_prod_string = f"in project '{forge_prod._store.bucket}' (Nexus env: '{forge_prod._store.endpoint}')"
    print(f"\n\nRetrieving Resource for '{res_name}' (Nexus id: '{res_id}', at tag '{res_tag}) {nexus_prod_string}")
    res_prod = forge_prod.retrieve(res_id, version=res_tag)
    if res_prod:
        print(f"\tFound Resource with id '{res_id}' and tag '{res_tag}' {nexus_prod_string})")
        print("\tNo synchronization will be performed")
        return

    print(f"\tNo Resource with id '{res_id}' and tag '{res_tag}' found {nexus_prod_string}")
    print(f"\tLooking for Resource with id '{res_id}' (dropping tag requirement)")
    res_prod = forge_prod.retrieve(res_id)
    if not res_prod:
        print(f"\tNo Resource with id '{res_id}' found {nexus_prod_string}")
        print(f"\tRegistering a new Resource for '{res_name}':")
        res_prod = Resource(id=res_id)
        forge_prod.register(res_prod)
    else:
        print(f"Resource found")

    print(f"\tSynchronizing Resource for '{res_name}' with staging version:")
    forge_staging = KnowledgeGraphForge("forge-config.yml",
        bucket=forge_prod._store.bucket, endpoint=staging_env, token=nexus_token_staging)
    nexus_staging_string = f"in project '{forge_prod._store.bucket}' (Nexus env: '{forge_staging._store.endpoint}'"
    print(f"\t\tRetrieving Resource for '{res_name}' (Nexus id: '{res_id}', at tag '{res_tag}) {nexus_staging_string}")
    res_staging = forge_staging.retrieve(res_id, version=res_tag)
    if not res_staging:
        if res_name in ["brain_realigned"]:
            print("\t\tIgnoring Resource")
            return
        else:
            raise Exception(f"\t\tNo Resource with id '{res_id}' and tag '{res_tag}' found {nexus_staging_string}")

    print("\t\tSynchronize Resource distribution")
    if not hasattr(res_staging.distribution.atLocation, "location"):
        raise Exception("The Resource.distribution.atLocation has no property 'location', maybe the Resource hsa not been migrated to gpfs?")
    distribution_file = res_staging.distribution.atLocation.location.replace("file:///gpfs", "/gpfs")
    res_prod.distribution = forge_prod.attach(distribution_file,
        content_type=res_staging.distribution.encodingFormat)

    print("\t\tSynchronize Resource attributes")
    skip_props = ["context", "id", "atlasRelease", "brainLocation", "distribution", "generation", "isRegisteredIn"]
    for attr, value in vars(res_staging).items():
        if attr.startswith("_") or attr in skip_props:
            continue
        setattr(res_prod, attr, value)

    print(f"\t\tUpdating Resource with id '{res_id}' {nexus_prod_string}")
    forge_prod.update(res_prod)
    print(f"\t\tTagging Resource with tag '{res_tag}' {nexus_prod_string}")
    forge_prod.tag(res_prod, res_tag)


def update_nexus_staging(nexus_ids, forge):
    file_nexus_map_path = os.path.join(metadata_dir, file_nexus_id_map)
    with open(file_nexus_map_path) as file_nexus_map_:
        file_nexus_map = json.loads(file_nexus_map_.read().strip())
    file_nexus_map_keys = list(file_nexus_map.keys())

    updated_map = False

    prob_maps = [str(path) for path in Path(metadata_dir).glob("probability_map_*.csv")]
    for prob_map_path in prob_maps:
        prob_map_doc_path = prob_map_path.replace(metadata_dir, os.path.join(metadata_dir, "docs")).replace(".csv", ".txt")
        if not os.path.isfile(prob_map_doc_path):
            raise Exception(f"No file {prob_map_doc_path} found, please provide it")
        with open(prob_map_doc_path) as prob_map_doc:
            prob_map_desc = prob_map_doc.read()

        nexus_id_path = nexus_ids[metadata_dir]

        prob_map = os.path.basename(prob_map_path)
        new_file = False
        nexus_id_key = None
        steps = None
        res_tag = None
        if prob_map not in file_nexus_map:
            print(f"File {prob_map} not found in map {file_nexus_map_path}, a new Resource will be created")
            new_file = True
            res = create_prob_map_resource(prob_map, prob_map_desc)
        else:
            file_nexus_map_keys.remove(prob_map)

            # Get Nexus ID
            nexus_id_key = file_nexus_map[prob_map]
            steps = nexus_id_key.split("/")
            # steps[0] is metadata_dir
            for step in steps[1:-1]:
                if step not in nexus_id_path:
                    raise Exception(f"No key '{step}' in {nexus_id_path} from {nexus_ids_path}")
                nexus_id_path = nexus_id_path[step]
            res_id_tag = nexus_id_path[steps[-1]]

            print(f"\nRetrieving Resource for {prob_map} (Nexus id: '{res_id_tag}'")
            res = forge.retrieve(res_id_tag)
            if not res:
                raise Exception(f"No Resource with id '{res_id_tag}' found in project '{forge._store.bucket}' (Nexus env: '{forge._store.endpoint}')")
            with open(prob_map_path, "rb") as prob_map_file:
                if res.distribution.digest.value == hashlib.sha256(prob_map_file.read()).hexdigest():
                    print(f"Hash of Resource distribution is identical to current file, nothing to update")
                    continue
            print(f"Hash of Resource distribution is different from hash of current file, updating the Resource")
            res.description = prob_map_desc

        res.distribution = forge.attach(prob_map_path, content_type="text/csv")
        if new_file:
            forge.register(res)
            new_tag = "v0.1.0"
            res_id_tag = res_tag_sep.join([res.id, new_tag])
            updated_map = True
        else:
            res._store_metadata._rev += 1  # current rev = tagged rev + 1
            forge.update(res)
            res_id, res_tag = res_id_tag.split(res_tag_sep)
            print(f"Increment minor version of tag '{res_tag}' for Resource id '{res_id}'")
            new_tag = increment_minor(res_tag)
        if not res._last_action.succeeded:
            raise Exception(f"The Resource registration/update failed with error:\n{res._last_action.message}")

        forge.tag(res, new_tag)
        if not res._last_action.succeeded:
            raise Exception(f"The Resource tagging failed with error:\n{res._last_action.message}")
        if new_file:
            print(f"Add entry of Resource {res_id_tag} in {nexus_ids_path}")
            prob_map_no_ext = prob_map.split(".")[0]
            nexus_id_path[prob_map_no_ext] = res_id_tag
            file_nexus_map[prob_map] = "/".join([metadata_dir, prob_map_no_ext])
        else:
            print(f"Update tag of Resource at '{nexus_id_key}' in {nexus_ids_path} to '{new_tag}'")
            nexus_id_path[steps[-1]] = res_id_tag.replace(res_tag, new_tag)

    if len(file_nexus_map_keys):
        print(f"The following elements from {file_nexus_map_path} are not found in {metadata_dir}: {', '.join(file_nexus_map_keys)}")

    with open(nexus_ids_path, 'w') as nexus_ids_file:
        nexus_ids_file.write(json.dumps(nexus_ids, indent=2))
    if updated_map:
        with open(file_nexus_map_path, 'w') as file_nexus_map_:
            file_nexus_map_.write(json.dumps(file_nexus_map, indent=2))


if __name__ == "__main__":
    synch_nexus()



from pathlib import Path
from kgforge.core import KnowledgeGraphForge
from cellCompVolume_payload import create_payload

with open(snakemake.log[0], "w") as logfile:
    total_densities = 0
    for folder in snakemake.params.input_paths:
        n_densities = len(list(Path(folder).rglob("*" + snakemake.params.files_ext)))
        logfile.write(f"\nAdding {n_densities} from {folder}")
        total_densities += n_densities
    logfile.write(f"\nExpecting {total_densities} densities in the CellCompositionVolume payload\n")

    forge = KnowledgeGraphForge(snakemake.params.forge_config, bucket=snakemake.params.nexus_bucket,
                                endpoint=snakemake.params.nexus_env, token=snakemake.params.nexus_token)

    logfile.write(
        f"Creating CellCompositionVolume payload for atlasRelease {snakemake.params.atlas_release_id} "
        f"with tag '{snakemake.params.resource_tag}' from endpoint '{snakemake.params.nexus_env}'\n")
    create_payload(forge, snakemake.params.atlas_release_id, snakemake.output.payload, total_densities,
        endpoint=snakemake.params.nexus_env, bucket=snakemake.params.nexus_bucket, tag=snakemake.params.resource_tag)
    logfile.write(f"CellCompositionVolume payload created: {snakemake.output.payload}\n")



'''
In the following, the term "preparing" means we locally initialize the resource,
including the creation of its ID, so that other resource can already establish a
link to it before it is even pushed.

Event should happen in this order:
- Preparing the new brain region ontology
- Preparing the new AtlasRelease locally
- Preparing the new annotation volume
- Preparing the (not really) new brain template volume

'''

import uuid
import argparse
from kgforge.core import KnowledgeGraphForge
from kgforge.core import Resource
import sys
import json
import os
import numpy as np
import nrrd
from datetime import datetime
import blue_brain_atlas_web_exporter.TreeIndexer as TreeIndexer



NRRD_TYPES_TO_NUMPY = {
    "signed char": "int8",
    "int8": "int8",
    "int8_t": "int8",
    "uchar": "uint8",
    "unsigned char": "uint8",
    "uint8": "uint8",
    "uint8_t": "uint8",
    "short": "int16",
    "short int": "int16",
    "signed short": "int16",
    "signed short int": "int16",
    "int16": "int16",
    "int16_t": "int16",
    "ushort": "int16",
    "unsigned short": "uint16",
    "unsigned short int": "uint16",
    "uint16": "uint16",
    "uint16_t": "uint16",
    "int": "int32",
    "signed int": "int32",
    "int32": "int32",
    "int32_t": "int32",
    "uint": "uint32",
    "unsigned int": "uint32",
    "uint32": "uint32",
    "uint32_t": "uint32",
    "longlong": "int64",
    "long long": "int64",
    "long long int": "int64",
    "signed long long": "int64",
    "signed long long int": "int64",
    "int64": "int64",
    "int64_t": "int64",
    "ulonglong": "uint64",
    "unsigned long long": "uint64",
    "unsigned long long int": "uint64",
    "uint64": "uint64",
    "uint64_t": "uint64",
    "float": "float32",
    "double": "float64"
}


dataSampleModalities = {
  "PARCELLATION_ID": "parcellationId",
  "CELL_TYPE_ID": "cellTypeId",
  "COLOR_RGB": "colorRGB",
  "COLOR_RGBA": "colorRGBA",
  "EULER_ANGLE": "eulerAngle",
  "GRADIENT_2D": "gradient2D",
  "GRADIENT_3D": "gradient3D",
  "LUMINANCE": "luminance",
  "MARKER_INTENSITY": "markerIntensity",
  "MASK": "mask",
  "POSITION_2D": "position2D",
  "POSITION_3D": "position3D",
  "QUANTITY": "quantity",
  "QUATERNION": "quaternion",
}

dataSampleModalitiesNumberOfComponents = {
  "PARCELLATION_ID": 1,
  "CELL_TYPE_ID": 1,
  "COLOR_RGB": 3,
  "COLOR_RGBA": 4,
  "EULER_ANGLE": 3,
  "GRADIENT_2D": 2,
  "GRADIENT_3D": 3,
  "LUMINANCE": 1,
  "MARKER_INTENSITY": 1,
  "MASK": 1,
  "POSITION_2D": 2,
  "POSITION_3D": 3,
  "QUANTITY": 1,
  "QUATERNION": 4,
}

VOLUMETRICDATALAYER_SCHEMAS_ID = "https://neuroshapes.org/dash/volumetricdatalayer"
ONTOLOGY_SCHEMAS_ID = "https://neuroshapes.org/dash/ontology"
ATLASRELEASE_SCHEMAS_ID = "https://neuroshapes.org/dash/atlasrelease"

def generate_payload_volumetric_data_layer(filepath, id, name, description, type, atlas_release_id, srs_id, region_node, sample_modality, resolution):
    
    nrrd_header = nrrd.read_header(filepath)
    return {
        "@id": id,
        "@type": [
            "VolumetricDataLayer",
            type,
            "Dataset"
        ],
        "atlasRelease": {
            "@id": atlas_release_id
        },
        "brainLocation": {
            "atlasSpatialReferenceSystem": {
            "@id": srs_id,
            "@type": [
                "AtlasSpatialReferenceSystem",
                "BrainAtlasSpatialReferenceSystem"
            ]
            },
            "brainRegion": {
            "@id": f"mba:{region_node['id']}",
            "label": region_node["name"],
            }
        },
        
        "bufferEncoding": nrrd_header["encoding"],

        "componentEncoding": NRRD_TYPES_TO_NUMPY[nrrd_header["type"]],

        "contribution": {
            "@type": "Contribution",
            "agent": {
            "@id": "https://ror.org/02s376052",
            "@type": [
                "Organization",
                "Agent"
            ]
            }
        },

        "dataSampleModality": dataSampleModalities[sample_modality],

        "description": description,
        "dimension": [
            {
            "@type": "ComponentDimension",
            "name": dataSampleModalities[sample_modality],
            "size": dataSampleModalitiesNumberOfComponents[sample_modality],
            },
            {
            "@type": "SpaceDimension",
            "size": 528,
            "unitCode": "voxel"
            },
            {
            "@type": "SpaceDimension",
            "size": 320,
            "unitCode": "voxel"
            },
            {
            "@type": "SpaceDimension",
            "size": 456,
            "unitCode": "voxel"
            }
        ],
   
        "endianness": nrrd_header.get("endian", "little"),

        "fileExtension": "nrrd",
        "isRegisteredIn": {
            "@id": srs_id,
            "@type": [
            "AtlasSpatialReferenceSystem",
            "BrainAtlasSpatialReferenceSystem"
            ]
        },
        "name": name,
        "resolution": {
            "unitCode": "µm",
            "value": resolution
        },
        "sampleType": "label",
        "subject": {
            "@type": "Subject",
            "species": {
            "@id": "http://purl.obolibrary.org/obo/NCBITaxon_10090",
            "label": "Mus musculus"
            }
        },
        "worldMatrix": [
            resolution,
            0,
            0,
            0,
            0,
            resolution,
            0,
            0,
            0,
            0,
            resolution,
            0,
            0,
            0,
            0,
            1
        ]
        }


def generate_payload_region_mesh(filepath, id, name, description, atlas_release_id, srs_id, region_id,):
    return {
        "@type": [
            "Mesh",
            "BrainParcellationMesh"
        ],
        "atlasRelease": {
            "@id": atlas_release_id
        },
        "brainLocation": {
            "atlasSpatialReferenceSystem": {
                "@id": srs_id,
                "@type": [
                    "AtlasSpatialReferenceSystem",
                    "BrainAtlasSpatialReferenceSystem"
                ]
          },
          "brainRegion": {
              "@id": f"mba:{region_id}",
          }
        },
        "description": description,
        "isRegisteredIn": {
            "@id": srs_id,
            "@type": [
                "BrainAtlasSpatialReferenceSystem",
                "AtlasSpatialReferenceSystem"
            ]
        },
        "name": name,
        "spatialUnit": "µm",
        "subject": {
            "@type": "Subject",
            "species": {
                "@id": "http://purl.obolibrary.org/obo/NCBITaxon_10090",
                "label": "Mus musculus"
            }
        }
    }


def generate_payload_brain_region_ontology(id, name):
    return {
      "@id": id,
      "@type": [
        "Ontology",
        "Entity",
        "ParcellationOntology"
      ],
      "label": name,
      "name": name,
      "wasDerivedFrom": {
        "@id": "http://ontology.neuinfo.org/NIF/ttl/generated/parcellation/mbaslim.ttl"
      }
    }


def generate_payload_atlas_release(atlas_release_id, name, description, brain_template_id, brain_region_ontology_id, parcellation_volume_id, srs_id):
    return {
        "@id": atlas_release_id,
        "@type": [
            "AtlasRelease",
            "BrainAtlasRelease"
        ],
        "brainTemplateDataLayer": {
            "@id": brain_template_id,
            "@type": "BrainTemplateDataLayer"
        },
        "description": description,
        "name": name,
        "parcellationOntology": {
            "@id": brain_region_ontology_id,
            "@type": "ParcellationOntology"
        },
        "parcellationVolume": {
            "@id": parcellation_volume_id,
            "@type": "BrainParcellationDataLayer"
        },
        "releaseDate": {
            "@type": "xsd:date",
            "@value": datetime.today().strftime("%Y-%m-%d")
        },
        "spatialReferenceSystem": {
            "@id": srs_id,
            "@type": "AtlasSpatialReferenceSystem"
        },
        "subject": {
            "@type": "Subject",
            "species": {
                "@id": "http://purl.obolibrary.org/obo/NCBITaxon_10090",
                "label": "Mus musculus"
            }
        }
    }


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--forge-config",
        dest="forge_config",
        required=True,
        metavar="<FILE PATH>",
        help="The path to the Nexus Forge configuration file")

    parser.add_argument(
        "--nexus-env",
        dest="nexus_env",
        required=True,
        metavar="<URL>",
        help="URL to the Nexus environment")

    parser.add_argument(
        "--nexus-org",
        dest="nexus_org",
        required=True,
        metavar="<NAME>",
        help="Name of the Nexus organization to push data to")

    parser.add_argument(
        "--nexus-proj",
        dest="nexus_proj",
        required=True,
        metavar="<NAME>",
        help="Name of the Nexus project to push data to")

    parser.add_argument(
        "--access-token",
        dest="access_token",
        required=True,
        metavar="<NAME>",
        help="Access token (JWT) to push data to Nexus")
        
    parser.add_argument(
        "--nexus-id-aibs-ccf-srs",
        dest="nexus_id_aibs_ccf_srs",
        required=True,
        metavar="<NEXUS ID>",
        help="Nexus ID of the Spatial Reference System")

    parser.add_argument(
        "--hierarchy",
        dest="hierarchy",
        required=True,
        metavar="<FILE PATH>",
        help="The hierarchy JSON file, sometimes called 1.json")

    parser.add_argument(
        "--hierarchy-ld",
        dest="hierarchy_ld",
        required=True,
        metavar="<FILE PATH>",
        help="The hierarchy in JSON-LD format")

    parser.add_argument(
        "--annotation-volume",
        dest="annotation_volume",
        required=True,
        metavar="<FILE PATH>",
        help="The NRRD brain region annotation volume")

    parser.add_argument(
        "--template-volume",
        dest="template_volume",
        required=True,
        metavar="<FILE PATH>",
        help="The NRRD brain template volume")

    parser.add_argument(
        "--out-atlasrelease-id-file",
        dest="out_atlasrelease_id_file",
        required=False,
        metavar="<FILE PATH>",
        help="The file to write the ID of the newly created AtlasRelease Nexus Resource")

    return parser.parse_args()



def main():
    args = parse_args()

    forge = KnowledgeGraphForge(
        args.forge_config,
        endpoint=args.nexus_env,
        bucket=f"{args.nexus_org}/{args.nexus_proj}",
        token=args.access_token,
        debug=True,
    )

    forge._debug=True

    # Predefining the @ids of the AtlaRelease and the items being pointed at by the AtlasRelease
    atlas_release_id = forge.format("identifier", str(uuid.uuid4()))
    brain_ontology_id = forge.format("identifier", str(uuid.uuid4()))
    brain_annotation_volume_id = forge.format("identifier", str(uuid.uuid4()))
    brain_template_volume_id = forge.format("identifier", str(uuid.uuid4()))
    srs_id = args.nexus_id_aibs_ccf_srs

    print("ID of atlas release to push: ", atlas_release_id)
    print("ID of brain region ontology to push: ", brain_ontology_id)
    print("ID of annotation volume to push: ", brain_annotation_volume_id)
    print("ID of template volume to push: ", brain_template_volume_id)
    print("Using SRS: ", srs_id)

    # Reading the brain region ontology, to later being able to generate
    # better names that include brain region labels
    brain_onto_json = json.loads(open(args.hierarchy, "r").read())
    # sometimes, the 1.json has its content in a "msg" sub prop (the original version has).
    # and some other versions don't. Here we deal with both
    if "msg" in brain_onto_json:
        onto_flat_tree = TreeIndexer.flattenTree(brain_onto_json['msg'][0])
    else:
        onto_flat_tree = TreeIndexer.flattenTree(brain_onto_json)

    # finding the root node of the brain region ontology
    root_node = None
    for node_id in onto_flat_tree:
        node = onto_flat_tree[node_id]
        if len(node["_ascendants"]) == 0:
            root_node = node
            break

    # Generating the base payload for the brain region ontology
    print("Pushing to Nexus: Brain region ontology...")
    brain_ontology_payload = generate_payload_brain_region_ontology(
        brain_ontology_id,
        "BBP Mouse Brain region ontology",
    )
    nexus_distrib_hierarchy = forge.attach(args.hierarchy, content_type="application/json")
    nexus_distrib_hierarchy_ld = forge.attach(args.hierarchy_ld, content_type="application/ld+json")
    nexus_resource_hierarchy = Resource.from_json(brain_ontology_payload)
    nexus_resource_hierarchy.distribution = [nexus_distrib_hierarchy, nexus_distrib_hierarchy_ld]
    forge.register(nexus_resource_hierarchy, ONTOLOGY_SCHEMAS_ID)

    # Generating the base payload for the brain annotation volume
    print("Pushing to Nexus: annotation volume...")
    brain_annotation_volume_payload = generate_payload_volumetric_data_layer(args.annotation_volume, 
        brain_annotation_volume_id,
        "BBP Mouse Brain Annotation Volume, 25µm",
        "This raster volume contains the brain region annotation as IDs, including the separation of cortical layers 2 and 3.",
        "BrainParcellationDataLayer",
        atlas_release_id,
        srs_id,
        root_node,
        "PARCELLATION_ID",
        25,
    )
    nexus_resource_annotation_volume = Resource.from_json(brain_annotation_volume_payload)
    nexus_resource_annotation_volume.distribution = forge.attach(args.annotation_volume, content_type="application/nrrd")
    forge.register(nexus_resource_annotation_volume, schema_id=VOLUMETRICDATALAYER_SCHEMAS_ID)

    # Generating the base payload for the brain annotation volume
    print("Pushing to Nexus: template volume...")
    brain_template_payload = generate_payload_volumetric_data_layer(args.template_volume, 
        brain_template_volume_id,
        "BBP Mouse Brain Template Volume, 25µm",
        "Raster volume of the brain template. This originaly comes from AIBS CCF (25µm)",
        "BrainTemplateDataLayer",
        atlas_release_id,
        srs_id,
        root_node,
        "LUMINANCE",
        25,
    )
    nexus_resource_template_volume = Resource.from_json(brain_template_payload)
    nexus_resource_template_volume.distribution = forge.attach(args.template_volume, content_type="application/nrrd")
    forge.register(nexus_resource_template_volume, schema_id=VOLUMETRICDATALAYER_SCHEMAS_ID)

    # Generating the base payload for the Atlas Release
    print("Pushing to Nexus: Atlas Release...")
    atlas_release_payload = generate_payload_atlas_release(atlas_release_id,
        "Blue Brain Atlas",
        "The official Atlas of the Blue Brain Project, derivated from AIBS Mouse CCF v3 (2017)",
        brain_template_volume_id,
        brain_ontology_id,
        brain_annotation_volume_id,
        srs_id
    )
    nexus_resource_atlas_release = Resource.from_json(atlas_release_payload)
    forge.register(nexus_resource_atlas_release, schema_id=ATLASRELEASE_SCHEMAS_ID)

    # Writing the text file that contains the ID of this atlasRelease
    open(args.out_atlasrelease_id_file, 'w').write(atlas_release_id)

if __name__ == "__main__":
    main()



import os

service_token_settings = os.environ["SERVICE_TOKEN_SETTINGS"]


def test_cli():
    working_dir = "tests/working_dir"
    rule_total = {"push_atlas_datasets": "63"}

    for rule, total in rule_total.items():
        cli_command = f"bbp-atlas --target-rule {rule} " \
            f"--snakemake-options '--config WORKING_DIR={working_dir} " \
            f"{service_token_settings}  -c1  --dryrun'"
        result = os.popen(cli_command).read()

        if any(x in result for x in ["Exception", "Invalid"]):
            print(result)
            assert False

        lines = result.splitlines()
        for line in lines:
            if line.startswith("total"):
                assert total in line, (f"Unexpected total number of rules for '{rule}' (expected: {total}):\n"
                                       f"{result}")



import voxcell
from atlas_commons.utils import assign_hemispheres

annotation = voxcell.VoxelData.load_nrrd(snakemake.input[0])
hemispheres = assign_hemispheres(annotation)
hemispheres.save_nrrd(snakemake.output[0])



import json

hasPart_key = "hasPart"


def create_payload(forge, atlas_release_id, output_file, n_densities_expected,
                   endpoint, bucket, tag=None):
    base_query = f"""
            ?s a METypeDensity ;
            atlasRelease <{atlas_release_id}>;
            brainLocation / brainRegion ?brainRegion ;
            distribution ?distribution ;
            _deprecated ?_deprecated;
            _project ?_project;
            """

    # Density Resources without annotation are not released in the CellCompositionVolume
    query_annotation = """
        SELECT DISTINCT ?s
        WHERE {""" + base_query + """
            annotation / hasBody / label ?mtype_label .
            ?distribution name ?nrrd_file ;
            contentUrl ?contentUrl .
            Filter (?_deprecated = 'false'^^xsd:boolean)
            Filter (?_project = <"""+endpoint+"""/projects/"""+bucket+""">)
        }"""
    all_resources_with_ann = forge.sparql(query_annotation, limit=3500, debug=False)
    print(f"{len(all_resources_with_ann)} ME-type densities with annotation found in total, filtering those with tag '{tag}'")

    resources = filter_by_tag(all_resources_with_ann, tag, forge)
    n_res_with_tag = len(resources)
    print(f"{n_res_with_tag} ME-type densities with annotation found with tag '{tag}'")
    assert n_res_with_tag == n_densities_expected, (f"The number of ME-type densities "
        f"found with tag '{tag}' ({n_res_with_tag}) does not match the expected number ({n_densities_expected})")

    # Get Generic{Excitatory,Inhibitory}Neuron
    for excInh in ["Excitatory", "Inhibitory"]:
        query_gen = f"""
            SELECT DISTINCT ?s
            WHERE {{""" + base_query + f"""
            annotation / hasBody <https://bbp.epfl.ch/ontologies/core/bmo/Generic{excInh}NeuronMType> ;
            annotation / hasBody <https://bbp.epfl.ch/ontologies/core/bmo/Generic{excInh}NeuronEType> .
            ?distribution name ?nrrd_file ;
            contentUrl ?contentUrl .
            Filter (?_deprecated = 'false'^^xsd:boolean)
            }}"""
        all_generic_resources = forge.sparql(query_gen, limit=1000, debug=False)
        generic_resources = filter_by_tag(all_generic_resources, tag, forge)
        assert len(generic_resources) == 1
        resources.extend(generic_resources)

    print(f"{len(resources)} ME-type densities will be released, including generic ones (tag '{tag}')")

    metype_annotations = [(a.hasBody for a in r.annotation) for r in resources] 

    mtype_to_etype = {}
    for i, metype_annotation_gen in enumerate(metype_annotations):
        metype_annotation_gen_list = list(metype_annotation_gen)
        if "MType" in metype_annotation_gen_list[0].type:
            if metype_annotation_gen_list[0].id not in mtype_to_etype:
                mtype_to_etype[metype_annotation_gen_list[0].id] = {"label": metype_annotation_gen_list[0].label}
            if "EType" in metype_annotation_gen_list[1].type and metype_annotation_gen_list[1].id not in mtype_to_etype[metype_annotation_gen_list[0].id]:
                mtype_to_etype[metype_annotation_gen_list[0].id][metype_annotation_gen_list[1].id] = {"label": metype_annotation_gen_list[1].label}
            if resources[i].id not in mtype_to_etype[metype_annotation_gen_list[0].id][metype_annotation_gen_list[1].id]:
                mtype_to_etype[metype_annotation_gen_list[0].id][metype_annotation_gen_list[1].id][resources[i].id] = {"type": resources[i].type, "_rev": resources[i]._store_metadata._rev}

    # CellCompositionVolume structure
    grouped_by_metype = {hasPart_key: []}
    for m_id, m in mtype_to_etype.items():
        m_content = {"@id": m_id, "label": m["label"], "about": ["https://neuroshapes.org/MType"], hasPart_key: []}
        for e_id, e in m.items():
            if e_id != "label":
                e_content = {"@id": e_id, "label": e["label"], "about": ["https://neuroshapes.org/EType"], hasPart_key: []}
                for res_id, res in e.items():
                    if res_id != "label":
                        e_content[hasPart_key].append({"@id": res_id, "@type": res["type"], "_rev": res["_rev"]})
                        m_content[hasPart_key].append(e_content)
        grouped_by_metype[hasPart_key].append(m_content)

    with open(output_file, "w") as f:
        json.dump(grouped_by_metype, f)

    return grouped_by_metype


def filter_by_tag(all_resources, tag, forge):
    if not tag:
        return all_resources

    tagged_resources = []
    for count, res in enumerate(all_resources):
        print(f"Retrieving Resource {count} of {len(all_resources)}")
        try:
            retrieved_res = forge.retrieve(id=res.s, version=tag)
            if retrieved_res is not None:
                tagged_resources.append(retrieved_res)
        except Exception:
            pass
    return tagged_resources



import os
import logging
import json
import yaml
from kgforge.core import KnowledgeGraphForge

from bba_data_push.push_atlas_release import validate_atlas_release

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

forge_config = "forge-config.yml"
nexus_token = os.environ["NEXUS_TOKEN"]
is_prod_env = os.environ.get("IS_PROD_ENV", False)
nexus_tag = os.environ["RESOURCE_TAG"]
if not nexus_tag:
    nexus_tag = "v1.1.0"

pipeline_config = yaml.safe_load(open("config.yaml").read().strip())
if is_prod_env:
    nexus_env = "prod"
    nexus_endpoint = pipeline_config["NEXUS_PROD_ENV"]
else:
    nexus_env = "staging"
    nexus_endpoint = pipeline_config["NEXUS_STAGING_ENV"]

nexus_ids = json.loads(open("nexus_ids.json").read().strip())
atlas_release_id = nexus_ids["AtlasRelease"][nexus_env]

logger.info(f"Validating AtlasRelease Id {atlas_release_id} at tag '{nexus_tag}' in Nexus {nexus_env} (IS_PROD_ENV: {is_prod_env})")
forge = KnowledgeGraphForge(forge_config, bucket="bbp/atlas", endpoint=nexus_endpoint, token=nexus_token)
validated = validate_atlas_release(atlas_release_id, forge, nexus_tag, logger)
if not validated:
    raise Exception(f"The properties of AtlasRelease Id {atlas_release_id} at "
                    f"tag '{nexus_tag}' did not pass the validation!")



#!/usr/bin/env python
from setuptools import setup, find_packages

with open("readme.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="bba_pipeline",
    author="Blue Brain Project, EPFL",
    use_scm_version=True,
    version="v1.3.0",
    setup_requires=["setuptools_scm"],
    description=(
        "Package defining a SnakeMake pipeline to create the Blue Brain Atlas datasets"
        "and push them into Nexus."
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/BlueBrain/bbp-atlas-pipeline",
    license="Apache-2.0",
    python_requires=">=3.9",
    install_requires=[
        "nexusforge>=0.8.2",
        "click>=7.0",
        "numpy>=1.19",
        "pynrrd>=0.4.0",
        "PyYAML>=5.3.1",
        "voxcell",
        "pulp==2.7.0",  # snakemake 7.32.3 breaks with pulp==2.8.0
        "snakemake==7.32.3",
        "blue-cwl",
        "python-gitlab"
    ],
    extras_require={
        "dev": ["pytest>=4.3", "pytest-cov>=2.8.0", "blue-brain-token-fetch>=1.0.0"],
        "docs": ["sphinx==7.1.2", "sphinx-bluebrain-theme", "myst-parser"]
    },
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        "console_scripts": ["bbp-atlas=bbp_atlas_cli:execute_pipeline"]
    },
)



import json
from kgforge.core import KnowledgeGraphForge
from blue_cwl import staging, statistics
import multiprocessing
import voxcell


with open(snakemake.log[0], "w") as logfile:
    logfile.write(f"Reading CellCompositionVolume payload from {snakemake.input.cellCompositionVolume}\n")
    with open(snakemake.input.cellCompositionVolume) as volume_json:
        volume_dict = json.load(volume_json)
        nexus_endpoint = snakemake.params.nexus_env
        forge = KnowledgeGraphForge(snakemake.params.forge_config, bucket=snakemake.params.nexus_bucket,
                                    endpoint=nexus_endpoint, token=snakemake.params.nexus_token)

        logfile.write(f"Creating density_distribution from endpoint '{nexus_endpoint}' in {snakemake.output.intermediate_density_distribution}\n")
        density_distribution = staging.materialize_density_distribution(forge=forge, dataset=volume_dict,
            output_file=snakemake.output.intermediate_density_distribution)

        logfile.write(f"Computing CellCompositionSummary payload for endpoint '{nexus_endpoint}'\n")

        with multiprocessing.Pool(processes=snakemake.params.cores) as pool:
            summary_statistics = statistics.atlas_densities_composition_summary(density_distribution,
                voxcell.RegionMap.load_json(snakemake.input.hierarchy),
                voxcell.VoxelData.load_nrrd(snakemake.input.annotation),
                map_function=pool.imap)
            logfile.write(f"Writing CellCompositionSummary payload in {snakemake.output.summary_statistics}\n")
            with open(snakemake.output.summary_statistics, "w") as outfile:
                outfile.write(json.dumps(summary_statistics, indent=4))



'''
This script should be launched after `push_atlasrelease.py` as it takes the ID
of the newly created atlas release in input.

This script is in charge of pushing the non-mask volume and volume that are not
directly involved in the composition of their atlasRelease (annotation + template), namely:
- placement hints volume
- direction vector volume
- orientation field volume

'''

import uuid
import argparse
import glob
from kgforge.core import KnowledgeGraphForge
from kgforge.core import Resource
import sys
import json
import os
import numpy as np
import nrrd
from datetime import datetime
import blue_brain_atlas_web_exporter.TreeIndexer as TreeIndexer



NRRD_TYPES_TO_NUMPY = {
    "signed char": "int8",
    "int8": "int8",
    "int8_t": "int8",
    "uchar": "uint8",
    "unsigned char": "uint8",
    "uint8": "uint8",
    "uint8_t": "uint8",
    "short": "int16",
    "short int": "int16",
    "signed short": "int16",
    "signed short int": "int16",
    "int16": "int16",
    "int16_t": "int16",
    "ushort": "int16",
    "unsigned short": "uint16",
    "unsigned short int": "uint16",
    "uint16": "uint16",
    "uint16_t": "uint16",
    "int": "int32",
    "signed int": "int32",
    "int32": "int32",
    "int32_t": "int32",
    "uint": "uint32",
    "unsigned int": "uint32",
    "uint32": "uint32",
    "uint32_t": "uint32",
    "longlong": "int64",
    "long long": "int64",
    "long long int": "int64",
    "signed long long": "int64",
    "signed long long int": "int64",
    "int64": "int64",
    "int64_t": "int64",
    "ulonglong": "uint64",
    "unsigned long long": "uint64",
    "unsigned long long int": "uint64",
    "uint64": "uint64",
    "uint64_t": "uint64",
    "float": "float32",
    "double": "float64"
}


dataSampleModalities = {
  "PARCELLATION_ID": "parcellationId",
  "CELL_TYPE_ID": "cellTypeId",
  "COLOR_RGB": "colorRGB",
  "COLOR_RGBA": "colorRGBA",
  "EULER_ANGLE": "eulerAngle",
  "GRADIENT_2D": "gradient2D",
  "GRADIENT_3D": "gradient3D",
  "VECTOR_3D": "vector3D",
  "LUMINANCE": "luminance",
  "MARKER_INTENSITY": "markerIntensity",
  "MASK": "mask",
  "POSITION_2D": "position2D",
  "POSITION_3D": "position3D",
  "QUANTITY": "quantity",
  "QUATERNION": "quaternion",
  "DISTANCE": "distance",
}

dataSampleModalitiesNumberOfComponents = {
  "PARCELLATION_ID": 1,
  "CELL_TYPE_ID": 1,
  "COLOR_RGB": 3,
  "COLOR_RGBA": 4,
  "EULER_ANGLE": 3,
  "GRADIENT_2D": 2,
  "GRADIENT_3D": 3,
  "VECTOR_3D": 3,
  "LUMINANCE": 1,
  "MARKER_INTENSITY": 1,
  "MASK": 1,
  "POSITION_2D": 2,
  "POSITION_3D": 3,
  "QUANTITY": 1,
  "QUATERNION": 4,
  "DISTANCE": 1,
}

VOLUMETRICDATALAYER_SCHEMAS_ID = "https://neuroshapes.org/dash/volumetricdatalayer"
ONTOLOGY_SCHEMAS_ID = "https://neuroshapes.org/dash/ontology"
ATLASRELEASE_SCHEMAS_ID = "https://neuroshapes.org/dash/atlasrelease"

def generate_payload_volumetric_data_layer(filepath, id, name, description, type, atlas_release_id, srs_id, region_node, sample_modality, resolution):
    
    nrrd_header = nrrd.read_header(filepath)
    return {
        "@id": id,
        "@type": [
            "VolumetricDataLayer",
            type,
            "Dataset"
        ],
        "atlasRelease": {
            "@id": atlas_release_id
        },
        "brainLocation": {
            "atlasSpatialReferenceSystem": {
            "@id": srs_id,
            "@type": [
                "AtlasSpatialReferenceSystem",
                "BrainAtlasSpatialReferenceSystem"
            ]
            },
            "brainRegion": {
            "@id": f"mba:{region_node['id']}",
            "label": region_node["name"],
            }
        },
        
        "bufferEncoding": nrrd_header["encoding"],

        "componentEncoding": NRRD_TYPES_TO_NUMPY[nrrd_header["type"]],

        "contribution": {
            "@type": "Contribution",
            "agent": {
            "@id": "https://ror.org/02s376052",
            "@type": [
                "Organization",
                "Agent"
            ]
            }
        },

        "dataSampleModality": dataSampleModalities[sample_modality],

        "description": description,
        "dimension": [
            {
            "@type": "ComponentDimension",
            "name": dataSampleModalities[sample_modality],
            "size": dataSampleModalitiesNumberOfComponents[sample_modality],
            },
            {
            "@type": "SpaceDimension",
            "size": 528,
            "unitCode": "voxel"
            },
            {
            "@type": "SpaceDimension",
            "size": 320,
            "unitCode": "voxel"
            },
            {
            "@type": "SpaceDimension",
            "size": 456,
            "unitCode": "voxel"
            }
        ],
   
        "endianness": nrrd_header.get("endian", "little"),

        "fileExtension": "nrrd",
        "isRegisteredIn": {
            "@id": srs_id,
            "@type": [
            "AtlasSpatialReferenceSystem",
            "BrainAtlasSpatialReferenceSystem"
            ]
        },
        "name": name,
        "resolution": {
            "unitCode": "µm",
            "value": resolution
        },
        "sampleType": "label",
        "subject": {
            "@type": "Subject",
            "species": {
            "@id": "http://purl.obolibrary.org/obo/NCBITaxon_10090",
            "label": "Mus musculus"
            }
        },
        "worldMatrix": [
            resolution,
            0,
            0,
            0,
            0,
            resolution,
            0,
            0,
            0,
            0,
            resolution,
            0,
            0,
            0,
            0,
            1
        ]
        }


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--forge-config",
        dest="forge_config",
        required=True,
        metavar="<FILE PATH>",
        help="The path to the Nexus Forge configuration file")

    parser.add_argument(
        "--nexus-env",
        dest="nexus_env",
        required=True,
        metavar="<URL>",
        help="URL to the Nexus environment")

    parser.add_argument(
        "--nexus-org",
        dest="nexus_org",
        required=True,
        metavar="<NAME>",
        help="Name of the Nexus organization to push data to")

    parser.add_argument(
        "--nexus-proj",
        dest="nexus_proj",
        required=True,
        metavar="<NAME>",
        help="Name of the Nexus project to push data to")

    parser.add_argument(
        "--access-token",
        dest="access_token",
        required=True,
        metavar="<NAME>",
        help="Access token (JWT) to push data to Nexus")
        
    parser.add_argument(
        "--nexus-id-aibs-ccf-srs",
        dest="nexus_id_aibs_ccf_srs",
        required=True,
        metavar="<NEXUS ID>",
        help="Nexus ID of the Spatial Reference System")

    parser.add_argument(
        "--hierarchy",
        dest="hierarchy",
        required=True,
        metavar="<FILE PATH>",
        help="The hierarchy JSON file, sometimes called 1.json")

    parser.add_argument(
        "--direction-vector-volume",
        dest="direction_vector_volume",
        required=True,
        metavar="<FILE PATH>",
        help="The NRRD direction vector volume (Eurler angles)")

    parser.add_argument(
        "--orientation-field-volume",
        dest="orientation_field_volume",
        required=True,
        metavar="<FILE PATH>",
        help="The NRRD orientation field volume (quaternions)")

    parser.add_argument(
        "--placement-hints-volume-dir",
        dest="placement_hints_volume_dir",
        required=True,
        metavar="<DIR PATH>",
        help="The NRRD placement hints volume directory")

    parser.add_argument(
        "--atlasrelease-id",
        dest="atlasrelease_id",
        required=False,
        metavar="<Nexus ID>",
        help="The ID of the AtlasRelease Nexus Resource the volumes are linked to")

    return parser.parse_args()



def main():
    args = parse_args()

    forge = KnowledgeGraphForge(
        args.forge_config,
        endpoint=args.nexus_env,
        bucket=f"{args.nexus_org}/{args.nexus_proj}",
        token=args.access_token,
        debug=True,
    )

    forge._debug=True

    # Predefining the @ids of the AtlaRelease and the items being pointed at by the AtlasRelease
    direction_vector_volume_id = forge.format("identifier", str(uuid.uuid4()))
    orientation_field_volume_id = forge.format("identifier", str(uuid.uuid4()))
    placement_hints_volume_id = forge.format("identifier", str(uuid.uuid4()))
    srs_id = args.nexus_id_aibs_ccf_srs
    atlas_release_id = args.atlasrelease_id

    print("ID of direction vector volume to push: ", direction_vector_volume_id)
    print("ID of orientation volume to push: ", orientation_field_volume_id)
    print("ID of placement hints volume to push: ", placement_hints_volume_id)
    print("Using SRS: ", srs_id)

    # Reading the brain region ontology, to later being able to generate
    # better names that include brain region labels
    brain_onto_json = json.loads(open(args.hierarchy, "r").read())
    # sometimes, the 1.json has its content in a "msg" sub prop (the original version has).
    # and some other versions don't. Here we deal with both
    if "msg" in brain_onto_json:
        onto_flat_tree = TreeIndexer.flattenTree(brain_onto_json['msg'][0])
    else:
        onto_flat_tree = TreeIndexer.flattenTree(brain_onto_json)

    # finding the root node of the brain region ontology
    root_node = None
    for node_id in onto_flat_tree:
        node = onto_flat_tree[node_id]
        if len(node["_ascendants"]) == 0:
            root_node = node
            break
    
    # Generating the base payload for the direction vector volume
    print("Pushing to Nexus: direction vector volume...")
    direction_vector_volume_payload = generate_payload_volumetric_data_layer(args.direction_vector_volume, 
        direction_vector_volume_id,
        "BBP Mouse Brain Direction Vector Volume, 25µm",
        "This raster volume contains the direction vectors as (x, y, z)",
        "CellOrientationField",
        atlas_release_id,
        srs_id,
        root_node,
        "VECTOR_3D",
        25,
    )
    nexus_resource_direction_vector_volume = Resource.from_json(direction_vector_volume_payload)
    nexus_resource_direction_vector_volume.distribution = forge.attach(args.direction_vector_volume, content_type="application/nrrd")
    forge.register(nexus_resource_direction_vector_volume, schema_id=VOLUMETRICDATALAYER_SCHEMAS_ID)

    # Generating the base payload for the orientation field (quaternion) volume
    print("Pushing to Nexus: orientation field volume...")
    orientation_field_volume_payload = generate_payload_volumetric_data_layer(args.orientation_field_volume , 
        orientation_field_volume_id,
        "BBP Mouse Brain Orientation Field Volume, 25µm",
        "This raster volume contains orientation field as quaternions",
        "CellOrientationField",
        atlas_release_id,
        srs_id,
        root_node,
        "QUATERNION",
        25,
    )
    nexus_resource_orientation_field_volume = Resource.from_json(orientation_field_volume_payload)
    nexus_resource_orientation_field_volume.distribution = forge.attach(args.orientation_field_volume, content_type="application/nrrd")
    forge.register(nexus_resource_orientation_field_volume, schema_id=VOLUMETRICDATALAYER_SCHEMAS_ID)

    # Generating the base payload for the placement hints volumes.
    # There are multiple nrrd files, one for each cortical layer and one for pia
    placement_hints_nrrd_paths = glob.glob(f"{args.placement_hints_volume_dir}/*.nrrd")
    print("Pushing to Nexus: placement hints volume...")
    placement_hints_volume_payload = generate_payload_volumetric_data_layer(placement_hints_nrrd_paths[0] , 
        placement_hints_volume_id,
        "BBP Mouse Brain Placement Hints Volumes, 25µm",
        "This raster volume contains placement hints volumes for all the cortical layers",
        "PlacementHintsDataLayer",
        atlas_release_id,
        srs_id,
        root_node,
        "DISTANCE",
        25,
    )
    nexus_resource_orientation_field_volume = Resource.from_json(placement_hints_volume_payload)
    placemement_hints_distributions = []
    for nrrd_path in placement_hints_nrrd_paths:
        placemement_hints_distributions.append(forge.attach(nrrd_path, content_type="application/nrrd"))

    # There is also a validation json file
    placement_hints_json_paths = glob.glob(f"{args.placement_hints_volume_dir}/*.json")
    for json_path in placement_hints_json_paths:
        placemement_hints_distributions.append(forge.attach(json_path, content_type="application/json"))

    nexus_resource_orientation_field_volume.distribution = placemement_hints_distributions
    forge.register(nexus_resource_orientation_field_volume, schema_id=VOLUMETRICDATALAYER_SCHEMAS_ID)


if __name__ == "__main__":
    main()



import os
import pytest
import numpy as np

from voxcell import RegionMap, VoxelData
from customize_pipeline.customize_pipeline import (merge_nrrd_files, get_region_id,
    check_rule_existence, get_var_path_map)


def test_merge_nrrd_files():
    test_folder = os.environ["TEST_FOLDER"]
    test_data = os.path.join(test_folder, "data")

    hierarchy_file = os.path.join(test_data, "hierarchy_leaves_only.json")
    annotation_file = os.path.join(test_data, "annotation_leaves_only.nrrd")
    default_output = os.path.join(test_data, "output.nrrd")
    merged_output_dir = os.path.join(test_data, "merged")
    region_volume_map = {
        315: os.path.join(test_data, "output_315"),
        549: os.path.join(test_data, "output_549"),
    }

    region_map = RegionMap.load_json(hierarchy_file)
    annotation = VoxelData.load_nrrd(annotation_file).raw
    results = merge_nrrd_files(region_map, annotation, region_volume_map,
                               default_output, merged_output_dir)

    assert len(results) == 1
    result = VoxelData.load_nrrd(results[0]).raw

    filename = os.path.basename(default_output)
    default_volume = VoxelData.load_nrrd(default_output)
    expected_volume = np.copy(default_volume.raw)
    for (region_id, volume_dir) in region_volume_map.items():
        ids_reg = region_map.find(region_id, "id", with_descendants=True)
        volume_file = os.path.join(volume_dir, filename)
        volume = VoxelData.load_nrrd(volume_file).raw
        # Get region mask
        region_mask = np.isin(annotation, list(ids_reg))
        # Supersede region {region_id} in result with values from volume
        expected_volume[region_mask] = volume[region_mask]

    assert result.shape == default_volume.shape
    assert np.array_equal(result, expected_volume)


def test_get_region_id():
    root_region = "http://api.brain-map.org/api/v2/data/Structure/997"
    region_id = get_region_id(root_region)
    assert region_id == 997


def test_check_rule_existence():
    rule_name = "test_rule_name"
    rule = None

    check_rule_existence(not rule, rule_name)
    with pytest.raises(Exception):
        check_rule_existence(rule, rule_name)


def test_get_var_path_map():
    available_vars = {
        "hierarchy": "HierarchyJson:hierarchy_ccfv3_l23split_barrelsplit",
        "annotation": "GeneratedDatasetPath:VolumetricFile:annotation_ccfv3_l23split_barrelsplit"}
    pipeline_dataset_config = {
        "HierarchyJson": {
            "hierarchy_ccfv3_l23split_barrelsplit": "hierarchy_path"},
        "GeneratedDatasetPath": {"VolumetricFile": {
            "annotation_ccfv3_l23split_barrelsplit": "annotation_path"}},
    }
    var_path_map_expected = {
        "hierarchy": "hierarchy_path",
        "annotation": "annotation_path"}

    var_path_map = get_var_path_map(available_vars, pipeline_dataset_config)
    assert var_path_map == var_path_map_expected
